"""
Read and write data files.
"""
import re
from collections import namedtuple

from .base import Reader
from .base import Root
from .base import Writer
from .google_api import GCellType
from .google_api import GDateTimeOption
from .google_api import GDrive
from .google_api import GMail
from .google_api import GSheets
from .google_api import GValueOption
from .readers import ExcelReader
from .readers import GSheetsReader
from .tables import extension_delimiter_map
from .tables import read_table_excel
from .tables import read_table_gsheets
from .tables import read_table_text
from .utils import _readers
from .utils import checksum
from .utils import copy
from .utils import git_head
from .utils import is_admin
from .utils import is_dir_accessible
from .utils import is_file_readable
from .utils import logger
from .utils import register
from .utils import remove_write_permissions
from .utils import run_as_admin
from .utils import search
from .utils import send_email
from .writers import HDF5Writer
from .writers import JSONWriter

__author__ = 'Measurement Standards Laboratory of New Zealand'
__copyright__ = '\xa9 2018 - 2023, ' + __author__
__version__ = '0.1.0'

_v = re.search(r'(\d+)\.(\d+)\.(\d+)[.-]?(.*)', __version__).groups()

version_info = namedtuple('version_info', 'major minor micro releaselevel')(int(_v[0]), int(_v[1]), int(_v[2]), 'final')
""":obj:`~collections.namedtuple`: Contains the version information as a (major, minor, micro, releaselevel) tuple."""


def read(file, **kwargs):
    """Read a file that has a :ref:`Reader <io-readers>` implemented.

    Parameters
    ----------
    file : :term:`path-like <path-like object>` or :term:`file-like <file object>`
        The file to read. For example, it could be a :class:`str` representing
        a file system path or a stream.
    **kwargs
        All keyword arguments are passed to the
        :meth:`Reader.can_read() <msl.io.base.Reader.can_read>`
        and :meth:`Reader.read() <msl.io.base.Reader.read>` methods.

    Returns
    -------
    :class:`~msl.io.base.Reader`
        The data from the file.

    Raises
    ------
    OSError
        If no :class:`~msl.io.base.Reader` exists to be able to read
        the specified file.
    """
    if hasattr(file, 'as_posix'):  # a pathlib.Path object
        file = str(file)

    if hasattr(file, 'read') or is_file_readable(file, strict=True):
        logger.debug('finding Reader for %r', file)
        for r in _readers:
            logger.debug('checking %s', r.__name__)
            try:
                can_read = r.can_read(file, **kwargs)
            except Exception as e:
                logger.debug('%s: %s [%s]', e.__class__.__name__, e, r.__name__)
                continue

            if can_read:
                logger.debug('reading file with %s', r.__name__)
                root = r(file)
                root.read(**kwargs)
                root.read_only = True
                return root

    raise OSError('No Reader exists to read {!r}'.format(file))


def read_table(file, **kwargs):
    """Read data in a table format from a file.

    A *table* has the following properties:

    1. The first row is a header.
    2. All rows have the same number of columns.
    3. All data values in a column have the same data type.

    Parameters
    ----------
    file : :term:`path-like <path-like object>` or :term:`file-like <file object>`
        The file to read. For example, it could be a :class:`str` representing
        a file system path or a stream. If `file` is a Google Sheets spreadsheet
        then `file` must end with ``.gsheet`` even if the ID of the spreadsheet
        is specified.
    **kwargs
        If the file is an Excel spreadsheet then the keyword arguments are passed to
        :func:`~msl.io.tables.read_table_excel`. If a Google Sheets spreadsheet then
        the keyword arguments are passed to :func:`~msl.io.tables.read_table_gsheets`.
        Otherwise, all keyword arguments are passed to :func:`~msl.io.tables.read_table_text`.

    Returns
    -------
    :class:`~msl.io.dataset.Dataset`
        The table as a :class:`~msl.io.dataset.Dataset`. The header is included as metadata.
    """
    extn = Reader.get_extension(file).lower()
    if extn.startswith('.xls'):
        return read_table_excel(file, **kwargs)
    elif extn == '.gsheet':
        if hasattr(file, 'as_posix'):  # a pathlib.Path object
            file = str(file)
        elif hasattr(file, 'name'):  # a TextIOWrapper object
            file = file.name
        return read_table_gsheets(file[:-7], **kwargs)  # ignore the extension
    else:
        return read_table_text(file, **kwargs)



"""
Read a data table from a file.
"""
import re

import numpy as np

from .base import Reader
from .dataset import Dataset
from .readers import ExcelReader
from .readers import GSheetsReader
from .utils import get_basename

_spreadsheet_top_left_regex = re.compile(r'^([A-Z]+)(\d+)$')
_spreadsheet_range_regex = re.compile(r'^[A-Z]+\d*:[A-Z]+\d*$')


extension_delimiter_map = {'.csv': ','}
""":class:`dict`: The delimiter to use to separate columns in a table based on the file extension.

If the `delimiter` is not specified when calling the :func:`~msl.io.read_table` function then this
extension-delimiter map is used to determine the value of the `delimiter`. If the file extension
is not in the map then the value of the `delimiter` is :data:`None` (i.e., split columns by any
whitespace). 

Examples
--------
You can customize your own map by adding key-value pairs

.. code-block:: pycon

    >>> from msl.io import extension_delimiter_map
    >>> extension_delimiter_map['.txt'] = '\\t'

"""


def read_table_text(file, **kwargs):
    """Read a data table from a text-based file.

    A *table* has the following properties:

    1. The first row is a header.
    2. All rows have the same number of columns.
    3. All data values in a column have the same data type.

    Parameters
    ----------
    file : :term:`path-like <path-like object>` or :term:`file-like <file object>`
        The file to read.
    **kwargs
        All keyword arguments are passed to :func:`~numpy.loadtxt`. If the
        `delimiter` is not specified and the `file` has ``csv`` as the file
        extension then the `delimiter` is automatically set to be ``','``.

    Returns
    -------
    :class:`~msl.io.dataset.Dataset`
        The table as a :class:`~msl.io.dataset.Dataset`. The header is included
        in the :class:`~msl.io.metadata.Metadata`.
    """
    if kwargs.get('unpack', False):
        raise ValueError('Cannot use the "unpack" option')

    if hasattr(file, 'as_posix'):  # a pathlib.Path object
        file = str(file)

    if 'delimiter' not in kwargs:
        extn = Reader.get_extension(file).lower()
        kwargs['delimiter'] = extension_delimiter_map.get(extn)

    if 'skiprows' not in kwargs:
        kwargs['skiprows'] = 0
    kwargs['skiprows'] += 1  # Reader.get_lines is 1-based, np.loadtxt is 0-based

    first_line = Reader.get_lines(file, kwargs['skiprows'], kwargs['skiprows'])
    if not first_line:
        header, data = [], []
    else:
        header = first_line[0].split(kwargs['delimiter'])
        # Calling np.loadtxt (on Python 3.5, 3.6 and 3.7) on a file
        # on a mapped drive could raise an OSError. This occurred
        # when a local folder was shared and then mapped on the same
        # computer. Opening the file using open() and then passing
        # in the file handle to np.loadtxt is more universal
        if hasattr(file, 'read'):  # already a file-like object
            data = np.loadtxt(file, **kwargs)
        else:
            with open(file, mode='rt') as fp:
                data = np.loadtxt(fp, **kwargs)
        use_cols = kwargs.get('usecols')
        if use_cols:
            if isinstance(use_cols, int):
                use_cols = [use_cols]
            header = [header[i] for i in use_cols]

    return Dataset(get_basename(file), None, True, data=data, header=np.asarray(header, dtype=str))


def read_table_excel(file, cells=None, sheet=None, as_datetime=True, dtype=None, **kwargs):
    """Read a data table from an Excel spreadsheet.

    A *table* has the following properties:

    1. The first row is a header.
    2. All rows have the same number of columns.
    3. All data values in a column have the same data type.

    Parameters
    ----------
    file : :term:`path-like <path-like object>` or :term:`file-like <file object>`
        The file to read.
    cells : :class:`str`, optional
        The cells to read. For example, ``C9`` will start at cell C9 and
        include all values until the end of the spreadsheet, ``A:C`` includes
        all rows in columns A, B and C, and, ``C9:G20`` includes
        values from only the specified cells. If not specified then returns
        all values from the specified `sheet`.
    sheet : :class:`str`, optional
        The name of the sheet to read the data from. If there is only one sheet
        in the workbook then you do not need to specify the name of the sheet.
    as_datetime : :class:`bool`, optional
        Whether dates should be returned as :class:`~datetime.datetime` or
        :class:`~datetime.date` objects. If :data:`False` then dates are returned
        as a :class:`str`.
    dtype : :class:`object`, optional
        If specified then it must be able to be converted to a :class:`~numpy.dtype` object.
    **kwargs
        All additional keyword arguments are passed to :func:`~xlrd.open_workbook`. Can use
        an `encoding` keyword argument as an alias for `encoding_override`.

    Returns
    -------
    :class:`~msl.io.dataset.Dataset`
        The table as a :class:`~msl.io.dataset.Dataset`. The header is included
        in the :class:`~msl.io.metadata.Metadata`.
    """
    if hasattr(file, 'as_posix'):  # a pathlib.Path object
        file = str(file)
    elif hasattr(file, 'name'):  # a TextIOWrapper object
        file = file.name

    excel = ExcelReader(file, **kwargs)

    if cells is not None and not _spreadsheet_range_regex.match(cells):
        match = _spreadsheet_top_left_regex.match(cells)
        if not match:
            raise ValueError('Invalid cell {!r}'.format(cells))
        name = sheet or excel.workbook.sheet_names()[0]
        s = excel.workbook.sheet_by_name(name)
        letters = excel.to_letters(s.ncols - 1)
        row = match.group(2)
        cells += ':{}{}'.format(letters, row)

    table = excel.read(cell=cells, sheet=sheet, as_datetime=as_datetime)
    excel.close()

    return _spreadsheet_to_dataset(table, file, dtype)


def read_table_gsheets(file, cells=None, sheet=None, as_datetime=True, dtype=None, **kwargs):
    """Read a data table from a Google Sheets spreadsheet.

    .. attention::
       You must have already performed the instructions specified in
       :class:`.GDrive` and in :class:`.GSheets` to be able to use this function.

    A *table* has the following properties:

    1. The first row is a header.
    2. All rows have the same number of columns.
    3. All data values in a column have the same data type.

    Parameters
    ----------
    file : :term:`path-like <path-like object>` or :term:`file-like <file object>`
        The file to read. Can be the ID of a Google Sheets spreadsheet.
    cells : :class:`str`, optional
        The cells to read. For example, ``C9`` will start at cell C9 and
        include all values until the end of the spreadsheet, ``A:C`` includes
        all rows in columns A, B and C, and, ``C9:G20`` includes
        values from only the specified cells. If not specified then returns
        all values from the specified `sheet`.
    sheet : :class:`str`, optional
        The name of the sheet to read the data from. If there is only one sheet
        in the spreadsheet then you do not need to specify the name of the sheet.
    as_datetime : :class:`bool`, optional
        Whether dates should be returned as :class:`~datetime.datetime` or
        :class:`~datetime.date` objects. If :data:`False` then dates are returned
        as a :class:`str`.
    dtype : :class:`object`, optional
        If specified then it must be able to be converted to a :class:`~numpy.dtype` object.
    **kwargs
        All additional keyword arguments are passed to :class:`~msl.io.readers.gsheets.GSheetsReader`.

    Returns
    -------
    :class:`~msl.io.dataset.Dataset`
        The table as a :class:`~msl.io.dataset.Dataset`. The header is included
        in the :class:`~msl.io.metadata.Metadata`.
    """
    if hasattr(file, 'as_posix'):  # a pathlib.Path object
        file = str(file)
    elif hasattr(file, 'name'):  # a TextIOWrapper object
        file = file.name

    sheets = GSheetsReader(file, **kwargs)

    if cells is not None and not _spreadsheet_range_regex.match(cells):
        if not _spreadsheet_top_left_regex.match(cells):
            raise ValueError('Invalid cell {!r}'.format(cells))
        r, c = sheets.to_indices(cells)
        data = sheets.read(sheet=sheet, as_datetime=as_datetime)
        table = [row[c:] for row in data[r:]]
    else:
        table = sheets.read(cell=cells, sheet=sheet, as_datetime=as_datetime)

    return _spreadsheet_to_dataset(table, file, dtype)


def _spreadsheet_to_dataset(table, file, dtype):
    if not table:
        header, data = [], []
    elif len(table) == 1:
        header, data = table[0], []
    elif len(table[0]) == 1:  # a single column
        header, data = table[0], [row[0] for row in table[1:]]
    else:
        header, data = table[0], table[1:]
        if len(data) == 1:  # a single row
            data = data[0]

    return Dataset(get_basename(file), None, True, data=data,
                   dtype=dtype, header=np.asarray(header, dtype=str))



import re

from msl.io import JSONWriter


def test_datasets_exclude():
    w = JSONWriter()
    assert len(list(w.datasets())) == 0

    w.create_dataset('pear', data=[1, 2, 3])  # ignore leading /
    w.create_dataset('/a/B/c/apple', data=[1, 2, 3])
    w.create_dataset('/a/strawberry', data=[1, 2, 3])
    w.create_dataset('a/b/banana', data=[1, 2, 3])  # ignore leading /
    w.create_dataset('/a/melon', data=[1, 2, 3])
    w.create_dataset('/a/B/c/d/e/kiwi', data=[1, 2, 3])

    # do not exclude datasets
    dsets = list(w.datasets())
    assert len(dsets) == 6
    assert dsets[0].name == '/pear'
    assert dsets[1].name == '/a/B/c/apple'
    assert dsets[2].name == '/a/strawberry'
    assert dsets[3].name == '/a/b/banana'
    assert dsets[4].name == '/a/melon'
    assert dsets[5].name == '/a/B/c/d/e/kiwi'

    # exclude all datasets with 'e' in the name
    dsets = list(w.datasets(exclude='e'))
    assert len(dsets) == 1
    assert dsets[0].name == '/a/b/banana'

    # only '/a/B/c/apple' should get excluded
    dsets = list(w.datasets(exclude='ApPl', flags=re.IGNORECASE))
    assert len(dsets) == 5
    assert dsets[0].name == '/pear'
    assert dsets[1].name == '/a/strawberry'
    assert dsets[2].name == '/a/b/banana'
    assert dsets[3].name == '/a/melon'
    assert dsets[4].name == '/a/B/c/d/e/kiwi'

    # everything in the '/a/B' Group should get excluded
    dsets = list(w.datasets(exclude='/a/B'))
    assert len(dsets) == 4
    assert dsets[0].name == '/pear'
    assert dsets[1].name == '/a/strawberry'
    assert dsets[2].name == '/a/b/banana'
    assert dsets[3].name == '/a/melon'

    # everything in the '/a/B' and '/a/b' Groups should get excluded
    dsets = list(w.datasets(exclude='/a/(B|b)'))
    assert len(dsets) == 3
    assert dsets[0].name == '/pear'
    assert dsets[1].name == '/a/strawberry'
    assert dsets[2].name == '/a/melon'

    # look at a subGroup
    dsets = list(w['/a/B/c'].datasets())
    assert len(dsets) == 2
    assert dsets[0].name == '/a/B/c/apple'
    assert dsets[1].name == '/a/B/c/d/e/kiwi'

    dsets = list(w.a.B.c.datasets())
    assert len(dsets) == 2
    assert dsets[0].name == '/a/B/c/apple'
    assert dsets[1].name == '/a/B/c/d/e/kiwi'

    # exclude everything with an 'e'
    dsets = list(w['/a/B/c'].datasets(exclude='e'))
    assert len(dsets) == 0

    dsets = list(w.a.B.c.datasets(exclude='e'))
    assert len(dsets) == 0

    # exclude everything in the 'e/' Group
    dsets = list(w['/a/B/c'].datasets(exclude='e/'))
    assert len(dsets) == 1
    assert dsets[0].name == '/a/B/c/apple'

    dsets = list(w.a.B.c.datasets(exclude='e/'))
    assert len(dsets) == 1
    assert dsets[0].name == '/a/B/c/apple'


def test_datasets_include():
    w = JSONWriter()
    assert len(list(w.datasets())) == 0

    w.create_dataset('pear', data=[1, 2, 3])  # ignore leading /
    w.create_dataset('a/B/c/apple', data=[1, 2, 3])  # ignore leading /
    w.create_dataset('/a/strawberry', data=[1, 2, 3])
    w.create_dataset('/a/b/banana', data=[1, 2, 3])
    w.create_dataset('/a/melon', data=[1, 2, 3])
    w.create_dataset('/a/B/c/d/e/kiwi', data=[1, 2, 3])

    # include all datasets
    dsets = list(w.datasets())
    assert len(dsets) == 6
    assert dsets[0].name == '/pear'
    assert dsets[1].name == '/a/B/c/apple'
    assert dsets[2].name == '/a/strawberry'
    assert dsets[3].name == '/a/b/banana'
    assert dsets[4].name == '/a/melon'
    assert dsets[5].name == '/a/B/c/d/e/kiwi'

    # include all datasets with 'e' in the name
    dsets = list(w.datasets(include='e'))
    assert len(dsets) == 5
    assert dsets[0].name == '/pear'
    assert dsets[1].name == '/a/B/c/apple'
    assert dsets[2].name == '/a/strawberry'
    assert dsets[3].name == '/a/melon'
    assert dsets[4].name == '/a/B/c/d/e/kiwi'

    # only '/a/B/c/apple' should get included
    dsets = list(w.datasets(include='ApPl', flags=re.IGNORECASE))
    assert len(dsets) == 1
    assert dsets[0].name == '/a/B/c/apple'

    # everything in the '/a/B' Group should get included
    dsets = list(w.datasets(include='/a/B'))
    assert len(dsets) == 2
    assert dsets[0].name == '/a/B/c/apple'
    assert dsets[1].name == '/a/B/c/d/e/kiwi'

    # everything in the '/a/B' and '/a/b' Groups should get included
    dsets = list(w.datasets(include='/a/(B|b)'))
    assert len(dsets) == 3
    assert dsets[0].name == '/a/B/c/apple'
    assert dsets[1].name == '/a/b/banana'
    assert dsets[2].name == '/a/B/c/d/e/kiwi'

    # look at a subGroup
    dsets = list(w['/a/B/c'].datasets())
    assert len(dsets) == 2
    assert dsets[0].name == '/a/B/c/apple'
    assert dsets[1].name == '/a/B/c/d/e/kiwi'

    dsets = list(w.a.B.c.datasets())
    assert len(dsets) == 2
    assert dsets[0].name == '/a/B/c/apple'
    assert dsets[1].name == '/a/B/c/d/e/kiwi'

    # include everything with 'pear'
    dsets = list(w['/a/B/c'].datasets(include='pear'))
    assert len(dsets) == 0

    dsets = list(w.a.B.c.datasets(include='pear'))
    assert len(dsets) == 0

    dsets = list(w.datasets(include='/pear'))
    assert len(dsets) == 1
    assert dsets[0].name == '/pear'

    # include everything in the 'e/' Group
    dsets = list(w['/a/B/c'].datasets(include='e/'))
    assert len(dsets) == 1
    assert dsets[0].name == '/a/B/c/d/e/kiwi'

    dsets = list(w.a.B.c.datasets(include='e/'))
    assert len(dsets) == 1
    assert dsets[0].name == '/a/B/c/d/e/kiwi'

    # exclude and include
    dsets = list(w['/a/B/c'].datasets(exclude='kiwi', include='e/'))
    assert len(dsets) == 0

    dsets = list(w.a.B.c.datasets(exclude='kiwi', include='e/'))
    assert len(dsets) == 0


def test_groups_exclude():
    w = JSONWriter()
    assert len(list(w.groups())) == 0

    w.create_group('pear')  # ignore leading /
    w.create_group('a/B/c/apple')  # ignore leading /
    w.create_group('/a/strawberry')
    w.create_group('/a/b/banana')
    w.create_group('/a/melon')
    w.create_group('/a/B/c/d/e/kiwi')

    # do not exclude any Groups
    groups = list(w.groups())
    assert len(groups) == 12
    assert groups[0].name == '/pear'
    assert groups[1].name == '/a'
    assert groups[2].name == '/a/B'
    assert groups[3].name == '/a/B/c'
    assert groups[4].name == '/a/B/c/apple'
    assert groups[5].name == '/a/strawberry'
    assert groups[6].name == '/a/b'
    assert groups[7].name == '/a/b/banana'
    assert groups[8].name == '/a/melon'
    assert groups[9].name == '/a/B/c/d'
    assert groups[10].name == '/a/B/c/d/e'
    assert groups[11].name == '/a/B/c/d/e/kiwi'

    # exclude all Groups with 'e' in the name
    groups = list(w.groups(exclude='e'))
    assert len(groups) == 6
    assert groups[0].name == '/a'
    assert groups[1].name == '/a/B'
    assert groups[2].name == '/a/B/c'
    assert groups[3].name == '/a/b'
    assert groups[4].name == '/a/b/banana'
    assert groups[5].name == '/a/B/c/d'

    # only '/a/B/c/apple' should get excluded
    groups = list(w.groups(exclude='ApPl', flags=re.IGNORECASE))
    assert len(groups) == 11
    assert groups[0].name == '/pear'
    assert groups[1].name == '/a'
    assert groups[2].name == '/a/B'
    assert groups[3].name == '/a/B/c'
    assert groups[4].name == '/a/strawberry'
    assert groups[5].name == '/a/b'
    assert groups[6].name == '/a/b/banana'
    assert groups[7].name == '/a/melon'
    assert groups[8].name == '/a/B/c/d'
    assert groups[9].name == '/a/B/c/d/e'
    assert groups[10].name == '/a/B/c/d/e/kiwi'

    # everything in the '/a/B' Group should get excluded
    groups = list(w.groups(exclude='/a/B'))
    assert len(groups) == 6
    assert groups[0].name == '/pear'
    assert groups[1].name == '/a'
    assert groups[2].name == '/a/strawberry'
    assert groups[3].name == '/a/b'
    assert groups[4].name == '/a/b/banana'
    assert groups[5].name == '/a/melon'

    # everything in the '/a/B' and '/a/b' Groups should get excluded
    groups = list(w.groups(exclude='/a/(B|b)'))
    assert len(groups) == 4
    assert groups[0].name == '/pear'
    assert groups[1].name == '/a'
    assert groups[2].name == '/a/strawberry'
    assert groups[3].name == '/a/melon'

    # look at a subGroup
    groups = list(w['/a/B/c'].groups())
    assert len(groups) == 4
    assert groups[0].name == '/a/B/c/apple'
    assert groups[1].name == '/a/B/c/d'
    assert groups[2].name == '/a/B/c/d/e'
    assert groups[3].name == '/a/B/c/d/e/kiwi'

    groups = list(w.a.B.c.groups())
    assert len(groups) == 4
    assert groups[0].name == '/a/B/c/apple'
    assert groups[1].name == '/a/B/c/d'
    assert groups[2].name == '/a/B/c/d/e'
    assert groups[3].name == '/a/B/c/d/e/kiwi'

    # exclude everything with an 'e'
    groups = list(w['/a/B/c'].groups(exclude='e'))
    assert len(groups) == 1
    assert groups[0].name == '/a/B/c/d'

    groups = list(w.a.B.c.groups(exclude='e'))
    assert len(groups) == 1
    assert groups[0].name == '/a/B/c/d'

    # exclude everything in the 'e/' Group
    groups = list(w['/a/B/c'].groups(exclude='e/'))
    assert len(groups) == 3
    assert groups[0].name == '/a/B/c/apple'
    assert groups[1].name == '/a/B/c/d'
    assert groups[2].name == '/a/B/c/d/e'

    groups = list(w.a.B.c.groups(exclude='e/'))
    assert len(groups) == 3
    assert groups[0].name == '/a/B/c/apple'
    assert groups[1].name == '/a/B/c/d'
    assert groups[2].name == '/a/B/c/d/e'


def test_groups_include():
    w = JSONWriter()
    assert len(list(w.groups())) == 0

    w.create_group('pear')  # ignore leading /
    w.create_group('a/B/c/apple')
    w.create_group('/a/strawberry')
    w.create_group('a/b/banana')  # ignore leading /
    w.create_group('/a/melon')
    w.create_group('/a/B/c/d/e/kiwi')

    # include all Groups
    groups = list(w.groups())
    assert len(groups) == 12
    assert groups[0].name == '/pear'
    assert groups[1].name == '/a'
    assert groups[2].name == '/a/B'
    assert groups[3].name == '/a/B/c'
    assert groups[4].name == '/a/B/c/apple'
    assert groups[5].name == '/a/strawberry'
    assert groups[6].name == '/a/b'
    assert groups[7].name == '/a/b/banana'
    assert groups[8].name == '/a/melon'
    assert groups[9].name == '/a/B/c/d'
    assert groups[10].name == '/a/B/c/d/e'
    assert groups[11].name == '/a/B/c/d/e/kiwi'

    # include all Groups with 'e' in the name
    groups = list(w.groups(include='e'))
    assert len(groups) == 6
    assert groups[0].name == '/pear'
    assert groups[1].name == '/a/B/c/apple'
    assert groups[2].name == '/a/strawberry'
    assert groups[3].name == '/a/melon'
    assert groups[4].name == '/a/B/c/d/e'
    assert groups[5].name == '/a/B/c/d/e/kiwi'

    # only '/a/B/c/apple' should get included
    groups = list(w.groups(include='ApPl', flags=re.IGNORECASE))
    assert len(groups) == 1
    assert groups[0].name == '/a/B/c/apple'

    # everything in the '/a/B' Group should get included
    groups = list(w.groups(include='/a/B'))
    assert len(groups) == 6
    assert groups[0].name == '/a/B'
    assert groups[1].name == '/a/B/c'
    assert groups[2].name == '/a/B/c/apple'
    assert groups[3].name == '/a/B/c/d'
    assert groups[4].name == '/a/B/c/d/e'
    assert groups[5].name == '/a/B/c/d/e/kiwi'

    # everything in the '/a/B' and '/a/b' Groups should get included
    groups = list(w.groups(include='/a/(B|b)'))
    assert len(groups) == 8
    assert groups[0].name == '/a/B'
    assert groups[1].name == '/a/B/c'
    assert groups[2].name == '/a/B/c/apple'
    assert groups[3].name == '/a/b'
    assert groups[4].name == '/a/b/banana'
    assert groups[5].name == '/a/B/c/d'
    assert groups[6].name == '/a/B/c/d/e'
    assert groups[7].name == '/a/B/c/d/e/kiwi'

    # look at a subGroup
    groups = list(w['/a/B/c'].groups())
    assert len(groups) == 4
    assert groups[0].name == '/a/B/c/apple'
    assert groups[1].name == '/a/B/c/d'
    assert groups[2].name == '/a/B/c/d/e'
    assert groups[3].name == '/a/B/c/d/e/kiwi'

    groups = list(w.a.B.c.groups())
    assert len(groups) == 4
    assert groups[0].name == '/a/B/c/apple'
    assert groups[1].name == '/a/B/c/d'
    assert groups[2].name == '/a/B/c/d/e'
    assert groups[3].name == '/a/B/c/d/e/kiwi'

    # include everything with an 'e'
    groups = list(w['/a/B/c'].groups(include='e'))
    assert len(groups) == 3
    assert groups[0].name == '/a/B/c/apple'
    assert groups[1].name == '/a/B/c/d/e'
    assert groups[2].name == '/a/B/c/d/e/kiwi'

    groups = list(w.a.B.c.groups(include='e'))
    assert len(groups) == 3
    assert groups[0].name == '/a/B/c/apple'
    assert groups[1].name == '/a/B/c/d/e'
    assert groups[2].name == '/a/B/c/d/e/kiwi'

    # include everything in the 'e/' Group
    groups = list(w['/a/B/c'].groups(include='e/'))
    assert len(groups) == 1
    assert groups[0].name == '/a/B/c/d/e/kiwi'

    groups = list(w.a.B.c.groups(include='e/'))
    assert len(groups) == 1
    assert groups[0].name == '/a/B/c/d/e/kiwi'

    # exclude and include
    groups = list(w['/a/B/c'].groups(exclude='apple', include='e'))
    assert len(groups) == 2
    assert groups[0].name == '/a/B/c/d/e'
    assert groups[1].name == '/a/B/c/d/e/kiwi'

    groups = list(w.a.B.c.groups(exclude='apple', include='e'))
    assert len(groups) == 2
    assert groups[0].name == '/a/B/c/d/e'
    assert groups[1].name == '/a/B/c/d/e/kiwi'


def test_ancestors():
    w = JSONWriter()
    assert len(list(w.ancestors())) == 0

    w.create_group('pear')  # ignore leading /
    w.create_group('a/B/c/apple')  # ignore leading /
    w.create_group('/a/strawberry')
    w.create_group('a/b/banana')  # ignore leading /
    w.create_group('/a/melon')
    w.create_group('/a/B/c/d/e/kiwi')

    ancestors = list(w.ancestors())
    assert len(ancestors) == 0

    ancestors = list(w.pear.ancestors())
    assert len(ancestors) == 1
    assert ancestors[0].name == '/'

    ancestors = list(w.a.ancestors())
    assert len(ancestors) == 1
    assert ancestors[0].name == '/'

    ancestors = list(w.a.B.ancestors())
    assert len(ancestors) == 2
    assert ancestors[0].name == '/a'
    assert ancestors[1].name == '/'

    ancestors = list(w.a.B.c.apple.ancestors())
    assert len(ancestors) == 4
    assert ancestors[0].name == '/a/B/c'
    assert ancestors[1].name == '/a/B'
    assert ancestors[2].name == '/a'
    assert ancestors[3].name == '/'

    ancestors = list(w['/a/B/c/d/e/kiwi'].ancestors())
    assert len(ancestors) == 6
    assert ancestors[0].name == '/a/B/c/d/e'
    assert ancestors[1].name == '/a/B/c/d'
    assert ancestors[2].name == '/a/B/c'
    assert ancestors[3].name == '/a/B'
    assert ancestors[4].name == '/a'
    assert ancestors[5].name == '/'

    ancestors = list(w['/a/B/c/d'].ancestors())
    assert len(ancestors) == 4
    assert ancestors[0].name == '/a/B/c'
    assert ancestors[1].name == '/a/B'
    assert ancestors[2].name == '/a'
    assert ancestors[3].name == '/'


def test_descendants():
    w = JSONWriter()
    assert len(list(w.descendants())) == 0

    w.create_group('pear')  # ignore leading /
    w.create_group('a/B/c/apple')  # ignore leading /
    w.create_group('/a/strawberry')
    w.create_group('/a/b/banana')
    w.create_group('/a/melon')
    w.create_group('/a/B/c/d/e/kiwi')

    descendants = list(w.descendants())
    assert len(descendants) == 12
    assert descendants[0].name == '/pear'
    assert descendants[1].name == '/a'
    assert descendants[2].name == '/a/B'
    assert descendants[3].name == '/a/B/c'
    assert descendants[4].name == '/a/B/c/apple'
    assert descendants[5].name == '/a/strawberry'
    assert descendants[6].name == '/a/b'
    assert descendants[7].name == '/a/b/banana'
    assert descendants[8].name == '/a/melon'
    assert descendants[9].name == '/a/B/c/d'
    assert descendants[10].name == '/a/B/c/d/e'
    assert descendants[11].name == '/a/B/c/d/e/kiwi'

    descendants = list(w.pear.descendants())
    assert len(descendants) == 0

    descendants = list(w.a.descendants())
    assert len(descendants) == 10
    assert descendants[0].name == '/a/B'
    assert descendants[1].name == '/a/B/c'
    assert descendants[2].name == '/a/B/c/apple'
    assert descendants[3].name == '/a/strawberry'
    assert descendants[4].name == '/a/b'
    assert descendants[5].name == '/a/b/banana'
    assert descendants[6].name == '/a/melon'
    assert descendants[7].name == '/a/B/c/d'
    assert descendants[8].name == '/a/B/c/d/e'
    assert descendants[9].name == '/a/B/c/d/e/kiwi'

    descendants = list(w.a.B.descendants())
    assert len(descendants) == 5
    assert descendants[0].name == '/a/B/c'
    assert descendants[1].name == '/a/B/c/apple'
    assert descendants[2].name == '/a/B/c/d'
    assert descendants[3].name == '/a/B/c/d/e'
    assert descendants[4].name == '/a/B/c/d/e/kiwi'

    descendants = list(w.a.B.c.apple.descendants())
    assert len(descendants) == 0

    descendants = list(w['/a/B/c'].descendants())
    assert len(descendants) == 4
    assert descendants[0].name == '/a/B/c/apple'
    assert descendants[1].name == '/a/B/c/d'
    assert descendants[2].name == '/a/B/c/d/e'
    assert descendants[3].name == '/a/B/c/d/e/kiwi'

    descendants = list(w['a/B/c/d'].descendants())  # ignore leading /
    assert len(descendants) == 2
    assert descendants[0].name == '/a/B/c/d/e'
    assert descendants[1].name == '/a/B/c/d/e/kiwi'



import logging
import os
import tempfile

import numpy as np
import pytest

try:
    import h5py
except ImportError:
    h5py = None

from msl.io import JSONWriter, HDF5Writer, read
from msl.io.dataset_logging import DatasetLogging

logger = logging.getLogger(__name__)

num_initial_handlers = 0


def setup_module():
    # Set the initial number of logging handlers.
    # Since pytest has its own this setup() function must be called
    # when pytest begins to test this module
    global num_initial_handlers
    num_initial_handlers = len(logging.getLogger().handlers)


def test_create():
    assert len(logging.getLogger().handlers) == num_initial_handlers

    root = JSONWriter()

    # also checks that specifying the level as a int is okay
    dset = root.create_dataset_logging('log', level=logging.INFO)

    assert dset.name == '/log'
    assert root.is_dataset(dset)
    assert isinstance(dset, DatasetLogging)
    assert len(logging.getLogger().handlers) == num_initial_handlers + 1
    assert len(dset) == 0
    assert np.array_equal(dset.dtype.names, ('asctime', 'levelname', 'name', 'message'))
    assert len(dset.metadata) == 3
    assert dset.metadata['logging_level'] == logging.INFO
    assert dset.metadata.logging_level_name == 'INFO'
    assert dset.metadata.logging_date_format == '%Y-%m-%dT%H:%M:%S.%f'

    logger.debug('hello')
    logger.info('world')
    logger.warning('foo')
    logger.error('bar')

    assert len(dset) == 3
    assert dset[ dset['levelname'] == 'ERROR']['message'] == 'bar'

    dset.remove_handler()
    assert len(logging.getLogger().handlers) == num_initial_handlers


def test_create_and_require():
    assert len(logging.getLogger().handlers) == num_initial_handlers

    root = JSONWriter()
    dset = root.create_dataset_logging('/a/b/log', level=logging.DEBUG)

    assert dset.name == '/a/b/log'
    assert len(logging.getLogger().handlers) == num_initial_handlers + 1

    with pytest.raises(ValueError):
        root.create_dataset_logging(dset.name)
    assert len(logging.getLogger().handlers) == num_initial_handlers + 1

    assert not dset.read_only
    assert root.is_dataset(dset)
    assert len(dset.metadata) == 3
    assert dset.metadata.logging_level == logging.DEBUG
    assert dset.metadata['logging_level_name'] == 'DEBUG'
    assert dset.metadata.logging_date_format == '%Y-%m-%dT%H:%M:%S.%f'

    messages = [
        'a debug message',
        'you should not do that so please be careful!',
        'tell me something useful',
        'NO!!!',
        'this is an error, cannot do that'
    ]

    logger.debug(messages[0])
    logger.warning(messages[1])
    logger.info(messages[2])
    logger.critical(messages[3])
    logger.error(messages[4])

    assert len(dset) == 5
    assert np.array_equal(dset['levelname'], ['DEBUG', 'WARNING', 'INFO', 'CRITICAL', 'ERROR'])
    assert np.array_equal(dset['message'], messages)

    b = root.a.b
    dset2 = b.require_dataset_logging('log')
    assert dset2 is dset

    logger.info('another info message')
    assert np.array_equal(dset['levelname'], ['DEBUG', 'WARNING', 'INFO', 'CRITICAL', 'ERROR', 'INFO'])
    assert np.array_equal(dset['message'], messages + ['another info message'])

    dset.remove_handler()
    assert len(logging.getLogger().handlers) == num_initial_handlers


def test_create_multiple_same_root():
    assert len(logging.getLogger().handlers) == num_initial_handlers

    root = JSONWriter()
    dset1 = root.create_dataset_logging('log')

    assert dset1.name == '/log'
    assert len(logging.getLogger().handlers) == num_initial_handlers + 1

    messages = [
        'a debug message',
        'you should not do that so please be careful!',
        'tell me something useful',
        'NO!!!',
        'this is an error, cannot do that'
    ]

    logger.debug(messages[0])
    logger.warning(messages[1])

    xx = root.create_group('xx')
    dset2 = xx.create_dataset_logging('log', level=logging.WARNING, attributes=['funcName', 'levelno'])
    assert dset2.name == '/xx/log'

    assert len(logging.getLogger().handlers) == num_initial_handlers + 2

    logger.info(messages[2])
    logger.critical(messages[3])
    logger.error(messages[4])

    assert dset1.level == logging.INFO
    assert len(dset1) == 4  # the DEBUG message is not there
    assert np.array_equal(dset1['levelname'], ['WARNING', 'INFO', 'CRITICAL', 'ERROR'])
    assert np.array_equal(dset1['message'], messages[1:])

    assert dset2.level == logging.WARNING
    assert len(dset2) == 2  # only ERROR and CRITICAL messages are there
    assert np.array_equal(dset2['levelno'], [logging.CRITICAL, logging.ERROR])
    assert np.array_equal(dset2['funcName'], ['test_create_multiple_same_root'] * 2)

    dset1.remove_handler()
    dset2.remove_handler()

    assert len(logging.getLogger().handlers) == num_initial_handlers


def test_requires_failures():
    assert len(logging.getLogger().handlers) == num_initial_handlers

    root = JSONWriter()
    root.create_dataset('regular')
    root.create_dataset_logging('logging')

    assert np.array_equal(root.logging.dtype.names, ['asctime', 'levelname', 'name', 'message'])
    with pytest.raises(ValueError, match=r"does not equal \('lineno', 'filename'\)"):
        root.require_dataset_logging('logging', attributes=['lineno', 'filename'])

    with pytest.raises(ValueError, match=r'not used for logging'):
        root.require_dataset_logging('regular')

    root.logging.remove_handler()
    assert len(logging.getLogger().handlers) == num_initial_handlers


def test_filter_loggers():
    assert len(logging.getLogger().handlers) == num_initial_handlers

    unlogger = logging.getLogger('unwanted')

    root = JSONWriter()
    dset = root.create_dataset_logging('log')
    dset.addFilter(logging.Filter(__name__))

    logger.info('ok')
    unlogger.info('not in dataset')

    assert len(dset) == 1
    assert dset['message'] == 'ok'

    dset.remove_handler()
    del logging.Logger.manager.loggerDict['unwanted']
    assert len(logging.getLogger().handlers) == num_initial_handlers


def test_save_then_read():
    assert len(logging.getLogger().handlers) == num_initial_handlers

    json = JSONWriter(file=os.path.join(tempfile.gettempdir(), 'msl-io-junk.json'))
    h5 = HDF5Writer(file=os.path.join(tempfile.gettempdir(), 'msl-io-junk.h5'))

    json.create_dataset_logging('log', date_fmt='%H:%M:%S', extra='ABC')
    h5.require_dataset_logging('/a/b/c/d/e/log')  # doesn't exist so creates it

    assert isinstance(json.log, DatasetLogging)
    assert isinstance(h5.a.b.c.d.e.log, DatasetLogging)

    logger.info('hello %s', 'world')
    logger.warning('foo')

    json.write(mode='w')
    if h5py is not None:
        h5.write(mode='w')

    json_2 = read(json.file)
    if h5py is not None:
        h5_2 = read(h5.file)
        os.remove(h5.file)
    os.remove(json.file)

    # when a file is read, what was once a DatasetLogging object is loaded as a regular Dataset
    # but can be turned back into a DatasetLogging by calling require_dataset_logging()
    assert not isinstance(json_2.log, DatasetLogging)
    if h5py is not None:
        assert not isinstance(h5_2.a.b.c.d.e.log, DatasetLogging)
    assert json_2.is_dataset(json_2.log)
    if h5py is not None:
        assert h5_2.is_dataset(h5_2.a.b.c.d.e.log)

    # convert the Dataset to DatasetLogging
    json_2.require_dataset_logging(json_2.log.name)
    if h5py is not None:
        h5_2.a.b.c.require_dataset_logging('/d/e/log')
    assert isinstance(json_2.log, DatasetLogging)
    if h5py is not None:
        assert isinstance(h5_2.a.b.c.d.e.log, DatasetLogging)
    assert json_2.is_dataset(json_2.log)
    if h5py is not None:
        assert h5_2.is_dataset(h5_2.a.b.c.d.e.log)

    assert len(json_2.log.metadata) == 4
    assert json_2.log.metadata['extra'] == 'ABC'
    assert json_2.log.metadata['logging_level'] == logging.INFO
    assert json_2.log.metadata.logging_level_name == 'INFO'
    assert json_2.log.metadata.logging_date_format == '%H:%M:%S'

    if h5py is not None:
        assert len(h5_2.a.b.c.d.e.log.metadata) == 3
        assert h5_2.a.b.c.d.e.log.metadata['logging_level'] == logging.INFO
        assert h5_2.a.b.c.d.e.log.metadata.logging_level_name == 'INFO'
        assert h5_2.a.b.c.d.e.log.metadata.logging_date_format == '%Y-%m-%dT%H:%M:%S.%f'

    assert np.array_equal(json_2.log['message'], ['hello world', 'foo'])
    if h5py is not None:
        if h5py.version.version_tuple.major < 3:
            assert np.array_equal(
                h5_2.a.b.c.d.e.log['message'],
                ['hello world', 'foo']
            )
        else:
            assert np.array_equal(
                h5_2.a.b.c.d.e.log['message'],
                [b'hello world', b'foo']
            )

    json.log.remove_handler()

    logger.info('baz')
    assert np.array_equal(json.log['message'], ['hello world', 'foo'])
    assert np.array_equal(h5.a.b.c.d.e.log['message'], ['hello world', 'foo', 'baz'])
    assert np.array_equal(json_2.log['message'], ['hello world', 'foo', 'baz'])
    if h5py is not None:
        if h5py.version.version_tuple.major < 3:
            assert np.array_equal(
                h5_2.a.b.c.d.e.log['message'],
                ['hello world', 'foo', 'baz']
            )
        else:
            assert np.array_equal(
                h5_2.a.b.c.d.e.log['message'].tolist(),
                [b'hello world', b'foo', 'baz']
            )

    h5.a.b.c.d.e.log.remove_handler()

    logger.warning('ooops...')
    logger.error('YIKES!')
    assert np.array_equal(json.log['message'], ['hello world', 'foo'])
    assert np.array_equal(h5.a.b.c.d.e.log['message'], ['hello world', 'foo', 'baz'])
    assert np.array_equal(json_2.log['message'], ['hello world', 'foo', 'baz', 'ooops...', 'YIKES!'])
    if h5py is not None:
        if h5py.version.version_tuple.major < 3:
            assert np.array_equal(
                h5_2.a.b.c.d.e.log['message'],
                ['hello world', 'foo', 'baz', 'ooops...', 'YIKES!']
            )
        else:
            assert np.array_equal(
                h5_2.a.b.c.d.e.log['message'].tolist(),
                [b'hello world', b'foo', 'baz', 'ooops...', 'YIKES!']
            )

    json_2.log.remove_handler()
    if h5py is not None:
        h5_2.a.b.c.d.e.log.remove_handler()

    assert len(logging.getLogger().handlers) == num_initial_handlers


def test_all_attributes():
    assert len(logging.getLogger().handlers) == num_initial_handlers

    attributes = [
        'asctime', 'created', 'filename', 'funcName', 'levelname',
        'levelno', 'lineno', 'message', 'module', 'msecs', 'name', 'pathname',
        'process', 'processName', 'relativeCreated', 'thread', 'threadName'
    ]

    json = JSONWriter()

    c = json.create_group('a/b/c')
    dset = c.create_dataset_logging('/d/e/f/log', level='DEBUG', attributes=attributes)

    assert dset.name == '/a/b/c/d/e/f/log'

    e = json.a.b.c.d.e
    dset2 = e.require_dataset_logging('/f/log')
    assert dset2 is dset

    logger.debug('d e b u g')
    logger.info('i n f o')
    logger.warning('w a r n i n g')
    logger.error('e r r o r')
    logger.critical('c r i t i c a l')

    assert len(dset) == 5
    assert len(dset[dset['levelno'] > logging.DEBUG]) == 4
    assert len(dset[dset['levelno'] > logging.INFO]) == 3
    assert len(dset[dset['levelno'] > logging.WARNING]) == 2
    assert len(dset[dset['levelno'] > logging.ERROR]) == 1
    assert len(dset[dset['levelno'] > logging.CRITICAL]) == 0
    assert np.array_equal(dset.dtype.names, attributes)

    dset.remove_handler()

    assert len(logging.getLogger().handlers) == num_initial_handlers


def test_is_logging_dataset():
    assert len(logging.getLogger().handlers) == num_initial_handlers

    root = JSONWriter()
    root.create_dataset('/a/b/regular')
    root.create_dataset_logging('log')
    root.create_dataset('regular')
    root.create_dataset_logging('/a/b/log')
    root.create_dataset_logging('/a/b/log2')

    log_dsets = [dset for dset in root.datasets() if root.is_dataset_logging(dset)]

    assert len(list(root.items())) == 7
    assert len(list(root.descendants())) == 2
    assert len(list(root.datasets())) == 5
    assert len(log_dsets) == 3

    assert len(logging.getLogger().handlers) == num_initial_handlers + 3

    for dset in root.datasets():
        if root.is_dataset_logging(dset):
            dset.remove_handler()

    assert len(logging.getLogger().handlers) == num_initial_handlers


def test_invalid_attributes():
    assert len(logging.getLogger().handlers) == num_initial_handlers

    root = JSONWriter()

    # cannot be an empty list/tuple
    with pytest.raises(ValueError):
        root.create_dataset_logging('log', attributes=[])
    with pytest.raises(ValueError):
        root.create_dataset_logging('log', attributes=tuple())

    # every element must be a string
    with pytest.raises(ValueError):
        root.create_dataset_logging('log', attributes=[1, 2, 3])
    with pytest.raises(ValueError):
        root.create_dataset_logging('log', attributes=['1', '2', 3])

    assert len(logging.getLogger().handlers) == num_initial_handlers


def test_initial_shape():
    assert len(logging.getLogger().handlers) == num_initial_handlers

    root = JSONWriter()

    num_records = 1234

    log1 = root.create_dataset_logging('log1', shape=(10000,))
    log2 = root.create_dataset_logging('log2')
    log3 = root.create_dataset_logging('log3', size=256)
    log4 = root.create_dataset_logging('log4', size=0)

    # the shape is an argument of Dataset and does not get passed to the Metadata
    assert 'shape' not in log1.metadata

    # specifying the `size` gets popped from the kwarg and gets converted to a `shape` kwarg
    assert 'size' not in log3.metadata
    assert 'shape' not in log3.metadata

    assert len(logging.getLogger().handlers) == num_initial_handlers + 4

    assert len(log1) == 10000
    assert len(log2) == 0
    assert len(log3) == 256
    assert len(log4) == 0

    for i in range(num_records):
        logging.info(i)  # just to be different, use the root logger

    assert len(log1) == 10000
    assert len(log2) == num_records
    assert len(log3) == 1380
    assert len(log4) == 1267

    for dset in root.datasets():
        dset.remove_empty_rows()

    assert len(log1) == num_records
    assert len(log2) == num_records
    assert len(log3) == num_records
    assert len(log4) == num_records

    for i in range(num_records):
        assert int(log1[i][3]) == i
        assert int(log2[i][3]) == i
        assert int(log3[i][3]) == i
        assert int(log4[i][3]) == i

    for dset in root.datasets():
        dset.remove_handler()

    assert len(logging.getLogger().handlers) == num_initial_handlers


def test_initial_index_value():
    assert len(logging.getLogger().handlers) == num_initial_handlers

    root = JSONWriter(file=os.path.join(tempfile.gettempdir(), 'msl-io-junk.json'))
    root.create_dataset_logging('log')

    n = 10

    for i in range(n):
        logger.info('message %d', i)

    assert len(root.log) == n

    root.write(mode='w')

    root2 = read(root.file)
    root3 = read(root.file)

    # specify more than n elements
    root2.require_dataset_logging(root.log.name, size=n+5)

    # specify less than n elements which automatically gets increased to n
    # also specify shape as an integer which gets cast to a 1-d tuple
    root3.require_dataset_logging(root.log.name, shape=n-5)

    assert len(logging.getLogger().handlers) == num_initial_handlers + 3

    os.remove(root.file)

    assert root2.log.size == n+5
    assert root3.log.size == n  # gets increased to n

    root.log.remove_handler()
    for i in range(n, 2*n):
        logger.info('message %d', i)

    assert root.log.size == n
    assert root2.log.size == 24
    assert root3.log.size == 27

    root.log.remove_empty_rows()
    root2.log.remove_empty_rows()
    root3.log.remove_empty_rows()

    assert root.log.size == n
    assert root2.log.size == 2*n
    assert root3.log.size == 2*n

    root2.log.remove_handler()
    for i in range(2*n, 3*n):
        logger.info('message %d', i)

    assert root.log.size == n
    assert root2.log.size == 2*n
    assert root3.log.size == 39

    root.log.remove_empty_rows()
    root2.log.remove_empty_rows()
    root3.log.remove_empty_rows()

    assert root.log.size == n
    assert root2.log.size == 2*n
    assert root3.log.size == 3*n

    messages1 = root['log']['message']
    messages2 = root2['log']['message']
    messages3 = root3['log']['message']
    for i in range(3*n):
        if i < n:
            assert messages1[i] == 'message %d' % i
        if i < 2*n:
            assert messages2[i] == 'message %d' % i
        assert messages3[i] == 'message %d' % i

    root3['log'].remove_handler()

    assert len(logging.getLogger().handlers) == num_initial_handlers


def test_invalid_shape_or_size():
    root = JSONWriter()

    with pytest.raises(ValueError, match=r'Invalid shape'):
        root.create_dataset_logging('log', shape=())

    with pytest.raises(ValueError, match=r'Invalid shape'):
        root.create_dataset_logging('log', shape=[])

    with pytest.raises(ValueError, match=r'Invalid shape'):
        root.create_dataset_logging('log', shape=(10, 5))

    with pytest.raises(ValueError, match=r'Invalid shape'):
        root.create_dataset_logging('log', shape=(-1,))

    with pytest.raises(ValueError, match=r'Invalid shape'):
        root.create_dataset_logging('log', size=-1)

    assert len(logging.getLogger().handlers) == num_initial_handlers


def test_set_logger():
    assert len(logging.getLogger().handlers) == num_initial_handlers

    root = JSONWriter()
    root.create_dataset_logging('log')

    for obj in [None, 'no', JSONWriter, logging.INFO, logging.Formatter, logging.Handler]:
        with pytest.raises(TypeError, match=r'Must be a logging.Logger object'):
            root.log.set_logger(obj)

    root.log.set_logger(logger)
    root.log.remove_handler()
    assert len(logging.getLogger().handlers) == num_initial_handlers


def test_hash():
    assert len(logging.getLogger().handlers) == num_initial_handlers

    root = JSONWriter()
    root.create_dataset_logging('log')

    # just tests that a hash value exists, don't care about the actual value
    assert isinstance(hash(root.log), int)

    root.log.remove_handler()
    assert len(logging.getLogger().handlers) == num_initial_handlers



"""
Reader for the HDF5_ file format.

.. attention::
   This Reader loads the entire HDF5_ file in memory. If you need to use any of
   the more advanced features of an HDF5_ file, it is best to directly load
   the file using H5py_.

.. _HDF5: https://www.hdfgroup.org/
.. _H5py: https://www.h5py.org/
"""
import os

try:
    import h5py
except ImportError:
    h5py = None

from ..base import Reader
from ..utils import register


@register
class HDF5Reader(Reader):
    """Reader for the HDF5_ file format."""

    @staticmethod
    def can_read(file, **kwargs):
        """The HDF5_ file format has a standard signature_.

        The first 8 bytes are ``\\x89HDF\\r\\n\\x1a\\n``.

        .. _signature: https://support.hdfgroup.org/HDF5/doc/H5.format.html#Superblock
        """
        return Reader.get_bytes(file, 8) == b'\x89HDF\r\n\x1a\n'

    def read(self, **kwargs):
        """Reads the HDF5_ file.

        Parameters
        ----------
        **kwargs
            All key-value pairs are passed to :class:`~h5py.File`.
        """
        if h5py is None:
            raise ImportError(
                'You must install h5py to read HDF5 files, run\n'
                '  pip install h5py'
            )

        def convert(name, obj):
            head, tail = os.path.split(name)
            s = self['/' + head] if head else self
            if isinstance(obj, h5py.Dataset):
                s.create_dataset(tail, data=obj[:], **obj.attrs)
            elif isinstance(obj, h5py.Group):
                s.create_group(tail, **obj.attrs)
            else:
                assert False, 'Unhandled HDF5Reader object {}'.format(obj)

        def h5_open(name):
            with h5py.File(name, mode='r', **kwargs) as h5:
                self.add_metadata(**h5.attrs)
                h5.visititems(convert)

        # Calling h5py.File on a file on a mapped drive could raise
        # an OSError. This occurred when a local folder was shared
        # and then mapped on the same computer. Opening the file
        # using open() and then passing in the file handle to
        # h5py.File is more universal
        if hasattr(self.file, 'read'):  # already a file-like object
            h5_open(self.file)
        else:
            with open(self.file, mode='rb') as fp:
                h5_open(fp)



"""
General functions.
"""
import ctypes
import hashlib
import logging
import os
import re
import shutil
import stat
import subprocess
import sys
from configparser import ConfigParser
from datetime import datetime
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from smtplib import SMTP

from .google_api import GMail

try:
    PermissionError
except NameError:
    PermissionError = OSError  # for Python 2.7
    FileExistsError = OSError

logger = logging.getLogger(__package__)

_readers = []


def checksum(file, algorithm='sha256', chunk_size=65536, shake_length=256):
    """Get the checksum of a file.

    A checksum is a sequence of numbers and letters that act as a fingerprint
    for a file against which later comparisons can be made to detect errors or
    changes in the file. It can be used to verify the integrity of the data.

    Parameters
    ----------
    file : :term:`path-like <path-like object>` or :term:`file <file object>` object
        A file to get the checksum of.
    algorithm : :class:`str`, optional
        The hash algorithm to use to compute the checksum.
        See :mod:`hashlib` for more details.
    chunk_size : :class:`int`, optional
        The number of bytes to read at a time from the file. It is useful
        to tweak this parameter when reading a large file to improve performance.
    shake_length : :class:`int`, optional
        The digest length to use for the ``SHAKE`` algorithm. See
        :meth:`hashlib.shake.hexdigest` for more details.

    Returns
    -------
    :class:`str`
        The checksum containing only hexadecimal digits.
    """
    def read(fp):
        # read in chucks in case the file size is too large
        # to load it into RAM all at once
        while True:
            data = fp.read(chunk_size)
            if not data:
                break
            hash_cls.update(data)

    hash_cls = hashlib.new(algorithm)

    try:
        with open(file, mode='rb') as f:
            read(f)
    except TypeError:
        if not hasattr(file, 'tell'):
            raise
        position = file.tell()
        read(file)
        file.seek(position)

    try:
        return hash_cls.hexdigest()
    except TypeError:
        return hash_cls.hexdigest(shake_length)


def copy(source, destination, overwrite=False, include_metadata=True):
    """Copy a file.

    Parameters
    ----------
    source : :term:`path-like object`
        The path to a file to copy.
    destination : :term:`path-like object`
        A directory to copy the file to or a full path (i.e., includes the basename).
        If the directory does not exist then it, and all intermediate directories,
        will be created.
    overwrite : :class:`bool`, optional
        Whether to overwrite the `destination` file if it already exists.
        If `destination` already exists and `overwrite` is :data:`False` then a
        :exc:`FileExistsError` is raised.
    include_metadata : :class:`bool`, optional
        Whether to also copy information such as the file permissions,
        the latest access time and latest modification time with the file.

    Returns
    -------
    :class:`str`
        The path to where the file was copied.
    """
    if os.path.isdir(destination) or is_dir_accessible(destination):
        destination = os.path.join(destination, os.path.basename(source))
    else:
        # TODO include the exist_ok kwarg to makedirs
        #  when dropping support for Python 2.7
        try:
            os.makedirs(os.path.dirname(destination))
        except OSError:
            pass

    if not overwrite and (os.path.isfile(destination) or is_file_readable(destination)):
        raise FileExistsError('Will not overwrite {!r}'.format(destination))

    # TODO include the follow_symlinks kwarg to copyfile and copystat
    #  (and to this "copy" function) when dropping support for Python 2.7
    shutil.copyfile(source, destination)
    if include_metadata:
        shutil.copystat(source, destination)

    return destination


def is_admin():
    """Check if the current process is being run as an administrator.

    Returns
    -------
    :class:`bool`
        Whether the current process is being run as an administrator.
    """
    try:
        return ctypes.windll.shell32.IsUserAnAdmin() == 1
    except AttributeError:
        try:
            return os.geteuid() == 0
        except AttributeError:
            return False


def is_dir_accessible(path, strict=False):
    """Check if a directory exists and is accessible.

    An accessible directory is one that the user has
    permission to access.

    Parameters
    ----------
    path : :class:`str`
        The directory to check.
    strict : :class:`bool`, optional
        Whether to raise the exception (if one occurs).

    Returns
    -------
    :class:`bool`
        Whether the directory exists and is accessible.
    """
    cwd = os.getcwd()
    try:
        os.chdir(path)
    except:
        if strict:
            raise
        return False
    else:
        os.chdir(cwd)
        return True


def is_file_readable(file, strict=False):
    """Check if a file exists and is readable.

    Parameters
    ----------
    file : :class:`str`
        The file to check.
    strict : :class:`bool`, optional
        Whether to raise the exception (if one occurs).

    Returns
    -------
    :class:`bool`
        Whether the file exists and is readable.
    """
    try:
        with open(file, mode='rb'):
            return True
    except:
        if strict:
            raise
        return False


def register(reader_class):
    """Use as a decorator to register a :class:`~msl.io.base.Reader` subclass.

    See :ref:`io-create-reader` for an example on how to use @register decorator.

    Parameters
    ----------
    reader_class : :class:`~msl.io.base.Reader`
        A :class:`~msl.io.base.Reader` subclass.

    Returns
    -------
    :class:`~msl.io.base.Reader`
        The :class:`~msl.io.base.Reader`.
    """
    def append(cls):
        _readers.append(cls)
        logger.debug('registered %r', cls)
        return cls
    return append(reader_class)


def search(folder, pattern=None, levels=0, regex_flags=0, exclude_folders=None,
           ignore_permission_error=True, ignore_hidden_folders=True, follow_symlinks=False):
    r"""Search for files starting from a root folder.

    Parameters
    ----------
    folder : :class:`str`
        The root folder to begin searching for files.
    pattern : :class:`str`, optional
        A regex string to use to filter the filenames. If :data:`None` then no
        filtering is applied and all files are yielded. Examples:

            * ``r'data'`` :math:`\rightarrow` find all files with the word ``data``
              in the filename

            * ``r'\.png$'`` :math:`\rightarrow` find all files with the extension ``.png``

            * ``r'\.jpe*g$'`` :math:`\rightarrow` find all files with the extension
              ``.jpeg`` or ``.jpg``

    levels : :class:`int`, optional
        The number of sub-folder levels to recursively search for files.
        If :data:`None` then search all sub-folders.
    regex_flags : :class:`int`, optional
        The flags to use to compile regex strings.
    exclude_folders : :class:`str` or :class:`list` of :class:`str`, optional
        The pattern of folder names to exclude from the search. Can be a regex
        string. If :data:`None` then include all folders in the search. Examples:

            * ``r'bin'`` :math:`\rightarrow` exclude all folders that contain the word ``bin``

            * ``r'^My'`` :math:`\rightarrow` exclude all folders that start with the letters ``My``

            * ``[r'bin', r'^My']`` which is equivalent to ``r'(bin|^My')`` :math:`\rightarrow` exclude
              all folders that contain the word ``bin`` or start with the letters ``My``

    ignore_permission_error : :class:`bool`, optional
        Whether to ignore :exc:`PermissionError` exceptions when reading
        the items within a folder.
    ignore_hidden_folders : :class:`bool`, optional
        Whether to ignore hidden folders from the search. A hidden folder
        starts with a ``.`` (a dot).
    follow_symlinks : :class:`bool`, optional
        Whether to search for files by following symbolic links.

    Yields
    ------
    :class:`str`
        The path to a file.
    """
    if levels is not None and levels < 0:
        return

    if ignore_hidden_folders and os.path.basename(folder).startswith('.'):
        logger.debug('ignore hidden folder %r', folder)
        return

    if exclude_folders:
        if isinstance(exclude_folders, str):
            exclude_folders = [exclude_folders]

        if isinstance(exclude_folders[0], str):
            ex_compiled = [re.compile(ex, flags=regex_flags) for ex in exclude_folders]
        else:  # the items should already be of type re.Pattern
            ex_compiled = exclude_folders

        basename = os.path.basename(folder)
        for exclude in ex_compiled:
            if exclude.search(basename):
                logger.debug('excluding folder %r', folder)
                return
    else:
        ex_compiled = None

    if ignore_permission_error:
        try:
            names = os.listdir(folder)
        except PermissionError:
            logger.debug('permission error %r', folder)
            return
    else:
        names = os.listdir(folder)

    if isinstance(pattern, str):
        regex = re.compile(pattern, flags=regex_flags) if pattern else None
    else:  # the value should already be of type re.Pattern
        regex = pattern

    for name in names:
        path = folder + '/' + name
        if os.path.isfile(path) or is_file_readable(path):
            if regex is None or regex.search(name):
                yield path
        elif os.path.isdir(path) or (follow_symlinks and os.path.islink(path)):
            for item in search(path,
                               pattern=regex,
                               levels=None if levels is None else levels - 1,
                               regex_flags=regex_flags,
                               exclude_folders=ex_compiled,
                               ignore_permission_error=ignore_permission_error,
                               ignore_hidden_folders=ignore_hidden_folders,
                               follow_symlinks=follow_symlinks):
                yield item


def send_email(config, recipients, sender=None, subject=None, body=None):
    """Send an email.

    Parameters
    ----------
    config
        A :term:`path-like object` or :term:`file-like object` of an INI-style
        configuration file that contains information on how to send an email.
        There are two ways to send an email -- Gmail API or SMTP server.

        An example INI file to use the Gmail API is the following (see
        :class:`~msl.io.google_api.GMail` for more details). Although all
        key-value pairs are optional, a ``[gmail]`` section must exist to use
        the Gmail API.

        .. code-block:: ini

           [gmail]
           account = work [default: None]
           credentials = path/to/client_secrets.json [default: None]
           scopes =       [default: None]
             https://www.googleapis.com/auth/gmail.send
             https://www.googleapis.com/auth/gmail.metadata
           domain = @gmail.com [default: None]

        An example INI file for an SMTP server is the following. Only the `host`
        and `port` key-value pairs are required.

        .. code-block:: ini

           [smtp]
           host = hostname or IP address of the SMTP server
           port = port number to connect to on the SMTP server
           starttls = true|yes|1|on -or- false|no|0|off [default: false]
           username = the username to authenticate with [default: None]
           password = the password for username [default: None]
           domain = @company.com [default: None]

        .. warning::
            Since this information is specified in plain text in the configuration
            file, you should set the file permissions provided by your operating
            system to ensure that your authentication credentials are safe.

    recipients : :class:`str` or :class:`list` of :class:`str`
        The email address(es) of the recipient(s). Can omit the ``@domain.com``
        part if a ``domain`` key is specified in the `config` file. Can be the
        value ``'me'`` if sending an email to yourself via Gmail.
    sender : :class:`str`, optional
        The email address of the sender. Can omit the ``@domain.com`` part
        if a ``domain`` key is specified in the `config` file. If not
        specified then it equals the value of the first `recipient` if using
        SMTP or the value ``'me'`` if using Gmail.
    subject : :class:`str`, optional
        The text to include in the subject field.
    body : :class:`str`, optional
        The text to include in the body of the email. The text can be
        enclosed in ``<html></html>`` tags to use HTML elements to format
        the message.
    """
    cfg = _prepare_email(config, recipients, sender)
    if cfg['type'] == 'smtp':
        server = SMTP(host=cfg['host'], port=cfg['port'])
        if cfg['starttls']:
            server.ehlo()
            server.starttls()
            server.ehlo()
        if cfg['username'] and cfg['password']:
            server.login(cfg['username'], cfg['password'])
        msg = MIMEMultipart()
        msg['From'] = cfg['from']
        msg['To'] = ', '.join(cfg['to'])
        msg['Subject'] = subject or '(no subject)'
        text = body or ''
        subtype = 'html' if text.startswith('<html>') else 'plain'
        msg.attach(MIMEText(text, subtype))
        server.sendmail(cfg['from'], cfg['to'], msg.as_string())
        server.quit()
    else:
        with GMail(account=cfg['account'], credentials=cfg['credentials'],
                   scopes=cfg['scopes']) as gmail:
            gmail.send(cfg['to'], sender=cfg['from'], subject=subject, body=body)


def _prepare_email(config, recipients, sender):
    """Loads a configuration file to prepare for sending an email.

    Returns a dict.
    """
    if hasattr(config, 'read'):
        contents = config.read()
    else:
        with open(config, mode='rt') as fp:
            contents = fp.read()

    if isinstance(contents, bytes):
        contents = contents.decode('utf-8')

    cp = ConfigParser()
    cp.read_string(contents)

    has_smtp = cp.has_section('smtp')
    has_gmail = cp.has_section('gmail')
    if has_smtp and has_gmail:
        raise ValueError("Cannot specify both a 'gmail' and 'smtp' section")
    if not (has_smtp or has_gmail):
        raise ValueError("Must create either a 'gmail' or 'smtp' section")

    section = cp['gmail'] if has_gmail else cp['smtp']

    domain = section.get('domain')
    if domain and not domain.startswith('@'):
        domain = '@' + domain

    if isinstance(recipients, str):
        recipients = [recipients]

    for i in range(len(recipients)):
        if domain and '@' not in recipients[i] and \
                (has_smtp or (has_gmail and recipients[i] != 'me')):
            recipients[i] += domain

    if not sender:
        if has_gmail:
            sender = 'me'
        else:
            sender = recipients[0]
    elif domain and ('@' not in sender) and \
            (has_smtp or (has_gmail and sender != 'me')):
        sender += domain

    cfg = {'type': section.name, 'to': recipients, 'from': sender}
    if has_smtp:
        host, port = section.get('host'), section.getint('port')
        if not (host and port):
            raise ValueError("Must specify the 'host' and 'port' of the SMTP server")

        username, password = section.get('username'), section.get('password')
        if username and not password:
            raise ValueError("Must specify the 'password' since a "
                             "'username' is specified")
        elif password and not username:
            raise ValueError("Must specify the 'username' since a "
                             "'password' is specified")

        cfg.update({
            'host': host,
            'port': port,
            'starttls': section.getboolean('starttls'),
            'username': username,
            'password': password,
        })
    else:
        scopes = section.get('scopes')
        cfg.update({
            'account': section.get('account'),
            'credentials': section.get('credentials'),
            'scopes': scopes.split() if scopes else None
        })
    return cfg


def get_basename(obj):
    """Get the :func:`~os.path.basename` of a file.

    Parameters
    ----------
    obj : :term:`path-like <path-like object>` or :term:`file-like <file object>`
        The object to get the :func:`~os.path.basename` of. If the object does not
        support the :func:`~os.path.basename` function then the
        :attr:`__name__ <definition.__name__>` of the `obj` is returned.

    Returns
    -------
    :class:`str`
        The basename of `obj`.
    """
    try:
        return os.path.basename(obj)
    except (TypeError, AttributeError):
        try:
            return os.path.basename(obj.name)
        except AttributeError:
            return obj.__class__.__name__


def git_head(directory):
    """Get information about the ``HEAD`` of a repository.

    This function requires that `git <https://git-scm.com/>`_ is installed
    and that it is available on ``PATH``.

    Parameters
    ----------
    directory : :class:`str`
        A directory that is under version control.

    Returns
    -------
    :class:`dict` or :data:`None`
        Information about the most recent commit on the current branch.
        If `directory` is not a directory that is under version control
        then returns :data:`None`.
    """
    cmd = ['git', 'show', '-s', '--format=%H %ct', 'HEAD']
    try:
        out = subprocess.check_output(cmd, cwd=directory, stderr=subprocess.PIPE)
    except subprocess.CalledProcessError:
        return None

    sha, timestamp = out.split()
    return {
        'hash': sha.decode('ascii'),
        'datetime': datetime.fromtimestamp(int(timestamp))
    }


def remove_write_permissions(path):
    """Remove all write permissions of a file.

    On Windows, this function will set the file attribute to be read only.

    On linux and macOS, write permission is removed for the User,
    Group and Others. The read and execute permissions are preserved.

    Parameters
    ----------
    path : :term:`path-like object`
        The path to remove the write permissions of.
    """
    current_permissions = stat.S_IMODE(os.lstat(path).st_mode)
    disable_writing = ~stat.S_IWUSR & ~stat.S_IWGRP & ~stat.S_IWOTH
    os.chmod(path, current_permissions & disable_writing)


def run_as_admin(args=None, executable=None, cwd=None, capture_stderr=False,
                 blocking=True, show=False, **kwargs):
    """Run a process as an administrator and return its output.

    Parameters
    ----------
    args : :class:`str` or :class:`list` of :class:`str`, optional
        A sequence of program arguments or else a single string. Providing a
        sequence of arguments is generally preferred, as it allows the module
        to take care of any required escaping and quoting of arguments
        (e.g., to permit spaces in file names).
    executable : :class:`str`, optional
        The executable to pass the `args` to.
    cwd : :class:`str`, optional
        The working directory for the elevated process.
    capture_stderr : :class:`bool`, optional
        Whether to send the stderr stream to stdout.
    blocking : :class:`bool`, optional
        Whether to wait for the process to finish before returning to the
        calling program.
    show : :class:`bool`, optional
        Whether to show the elevated console (Windows only). If
        :data:`True` then the stdout stream of the process is not captured.
    kwargs
        If the current process already has admin privileges or if the operating
        system is not Windows then all additional keyword arguments are passed
        to :func:`~subprocess.check_output`. Otherwise, only a `timeout` keyword
        argument is used (Windows).

    Returns
    -------
    :class:`bytes`, :class:`int` or :class:`~subprocess.Popen`
        The returned object depends on whether the process is executed in blocking
        or non-blocking mode. If blocking then :class:`bytes` are returned (the
        stdout stream of the process). If non-blocking, then the returned object
        will either be the :class:`~subprocess.Popen` instance that is running the
        process (POSIX) or an :class:`int` which is the process ID (Windows).

    Examples
    --------
    .. invisible-code-block: pycon

       >>> SKIP_RUN_AS_ADMIN()

    Import the modules

    >>> import sys
    >>> from msl.io import run_as_admin

    Run a shell script

    >>> run_as_admin(['./script.sh', '--message', 'hello world'])

    Run a Python script

    >>> run_as_admin([sys.executable, 'script.py', '--verbose'], cwd='D:\\\\My Scripts')

    Create a service in the Windows registry and in the Service Control Manager database

    >>> run_as_admin(['sc', 'create', 'MyLogger', 'binPath=', 'C:\\\\logger.exe', 'start=', 'auto'])
    """
    if not args and not executable:
        raise ValueError('Must specify the args and/or an executable')

    stderr = subprocess.STDOUT if capture_stderr else None
    process = subprocess.check_output if blocking else subprocess.Popen

    if is_admin():
        return process(args, executable=executable, cwd=cwd,
                       stderr=stderr, **kwargs)

    if cwd is None:
        cwd = os.getcwd()

    if os.name != 'nt':
        if not args:
            command = ['sudo', executable]
        elif isinstance(args, str):
            exe = executable or ''
            command = 'sudo {} {}'.format(exe, args)
        else:
            exe = [executable] if executable else []
            command = ['sudo'] + exe + list(args)
        return process(command, cwd=cwd, stderr=stderr, **kwargs)

    # Windows is more complicated

    if args is None:
        args = ''

    if not isinstance(args, str):
        args = subprocess.list2cmdline(args)

    if executable is None:
        executable = ''
    else:
        executable = subprocess.list2cmdline([executable])

    # the 'runas' verb starts in C:\WINDOWS\system32
    cd = subprocess.list2cmdline(['cd', '/d', cwd, '&&'])

    # check if a Python environment needs to be activated
    activate = ''
    if executable == sys.executable or args.startswith(sys.executable):
        conda = os.getenv('CONDA_PREFIX')  # conda
        venv = os.getenv('VIRTUAL_ENV')  # venv
        if conda:
            env = os.getenv('CONDA_DEFAULT_ENV')
            assert env, 'CONDA_DEFAULT_ENV environment variable does not exist'
            if env == 'base':
                bat = os.path.join(conda, 'Scripts', 'activate.bat')
            else:
                bat = os.path.abspath(os.path.join(conda, os.pardir, os.pardir,
                                                   'Scripts', 'activate.bat'))
            assert os.path.isfile(bat), 'Cannot find {!r}'.format(bat)
            activate = subprocess.list2cmdline([bat, env, '&&'])
        elif venv:
            bat = os.path.join(venv, 'Scripts', 'activate.bat')
            assert os.path.isfile(bat), 'Cannot find {!r}'.format(bat)
            activate = subprocess.list2cmdline([bat, '&&'])

    # redirect stdout (stderr) to a file
    redirect = ''
    stdout_file = ''
    if not show:
        import uuid
        import tempfile
        stdout_file = os.path.join(tempfile.gettempdir(), str(uuid.uuid4()))
        r = ['>', stdout_file]
        if capture_stderr:
            r.append('2>&1')
        redirect = subprocess.list2cmdline(r)
        if re.search(r'\d$', args):
            # this number is also considered as a file handle, so add a space
            redirect = ' ' + redirect

    # the string that is passed to cmd.exe
    params = '/S /C "{cd} {activate} {executable} {args}"{redirect}'.format(
        cd=cd, activate=activate, executable=executable, args=args, redirect=redirect)

    from ctypes.wintypes import DWORD, ULONG, HWND, LPCWSTR, INT, HINSTANCE, HKEY, HANDLE

    class ShellExecuteInfoW(ctypes.Structure):
        _fields_ = [
            ('cbSize', DWORD),
            ('fMask', ULONG),
            ('hwnd', HWND),
            ('lpVerb', LPCWSTR),
            ('lpFile', LPCWSTR),
            ('lpParameters', LPCWSTR),
            ('lpDirectory', LPCWSTR),
            ('nShow', INT),
            ('hInstApp', HINSTANCE),
            ('lpIDList', ctypes.c_void_p),
            ('lpClass', LPCWSTR),
            ('hkeyClass', HKEY),
            ('dwHotKey', DWORD),
            ('hIcon', HANDLE),
            ('hProcess', HANDLE)]

    sei = ShellExecuteInfoW()
    sei.fMask = 0x00000040 | 0x00008000  # SEE_MASK_NOCLOSEPROCESS | SEE_MASK_NO_CONSOLE
    sei.lpVerb = kwargs.get('verb', u'runas')  # change the verb when running the tests
    sei.lpFile = u'cmd.exe'
    sei.lpParameters = params
    sei.lpDirectory = u'{}'.format(cwd) if cwd else None
    sei.nShow = int(show)
    sei.cbSize = ctypes.sizeof(sei)
    if not ctypes.windll.Shell32.ShellExecuteExW(ctypes.byref(sei)):
        raise ctypes.WinError()

    if not blocking:
        return sei.hProcess

    kernel32 = ctypes.windll.kernel32
    timeout = kwargs.get('timeout', -1)  # INFINITE = -1
    milliseconds = int(timeout * 1e3) if timeout > 0 else timeout

    ret = kernel32.WaitForSingleObject(sei.hProcess, milliseconds)
    if ret == 0:  # WAIT_OBJECT_0
        stdout = b''
        if stdout_file and os.path.isfile(stdout_file):
            with open(stdout_file, mode='rb') as fp:
                stdout = fp.read()
            os.remove(stdout_file)

        code = DWORD()
        if not kernel32.GetExitCodeProcess(sei.hProcess, ctypes.byref(code)):
            raise ctypes.WinError()

        if code.value != 0:
            msg = ctypes.FormatError(code.value)
            out_str = stdout.decode('utf-8', 'ignore').rstrip()
            if show:
                msg += '\nSet show=False to capture the stdout stream.'
            else:
                if not capture_stderr:
                    msg += '\nSet capture_stderr=True to see if ' \
                           'more information is available.'
                if out_str:
                    msg += '\n{}'.format(out_str)
            raise ctypes.WinError(code=code.value, descr=msg)

        kernel32.CloseHandle(sei.hProcess)
        return stdout

    if ret == 0xFFFFFFFF:  # WAIT_FAILED
        raise ctypes.WinError()

    if ret == 0x00000080:  # WAIT_ABANDONED
        msg = 'The specified object is a mutex object that was not ' \
              'released by the thread that owned the mutex object before ' \
              'the owning thread terminated. Ownership of the mutex ' \
              'object is granted to the calling thread and the mutex state ' \
              'is set to non-signaled. If the mutex was protecting persistent ' \
              'state information, you should check it for consistency.'
    elif ret == 0x00000102:  # WAIT_TIMEOUT
        msg = "The timeout interval elapsed after {} second(s) and the " \
              "object's state is non-signaled.".format(timeout)
    else:
        msg = 'Unknown return value 0x{:x}'.format(ret)

    raise WindowsError('WaitForSingleObject: ' + msg)



import pytest

from msl.io.base import Root
from msl.io.base import Writer
from msl.io.dataset import Dataset


def test_set_root():
    root = Root('some file')

    writer = Writer()
    assert len(writer) == 0
    assert writer.file is None
    assert len(writer.metadata) == 0

    for item in [dict(), tuple(), list(), None, Dataset('dset', None, False)]:
        with pytest.raises(TypeError):
            writer.set_root(item)

    # set an empty root is okay
    writer.set_root(root)
    assert len(writer) == 0
    assert writer.file is None
    assert len(writer.metadata) == 0

    # set an empty root with metadata
    root = Root('some file', one=1, foo='bar')
    writer.set_root(root)
    assert len(writer) == 0
    assert writer.file is None
    assert len(writer.metadata) == 2
    assert writer.metadata.one == 1
    assert writer.metadata.foo == 'bar'

    # check that the metadata gets replaced
    writer = Writer('a filename', meta='data')
    assert len(writer) == 0
    assert writer.file == 'a filename'
    assert len(writer.metadata) == 1
    assert writer.metadata.meta == 'data'
    writer.set_root(root)
    assert len(writer.metadata) == 2
    assert writer.metadata.one == 1
    assert writer.metadata.foo == 'bar'
    assert 'meta' not in writer.metadata
    assert writer.file == 'a filename'  # the URL does not change

    a = root.create_group('a')
    a.create_dataset('d1')
    b = a.create_group('b')
    b.create_dataset('d2')
    c = b.create_group('c')
    root.create_group('x/y/z')
    d = c.create_group('d')
    a.create_dataset('d3')
    root.create_dataset('d4')
    d.create_dataset('d5')
    d.create_dataset('d6')
    root.create_dataset('d7')
    root.add_metadata(two=2, three=3)

    writer.set_root(root)

    assert len(writer) == 14
    assert len(list(writer.groups())) == 7
    assert len(list(writer.datasets())) == 7
    assert 'a' in writer
    assert 'd1' in writer.a
    assert 'b' in writer.a
    assert 'd2' in writer.a.b
    assert 'c' in writer.a.b
    assert 'x' in writer
    assert 'x/y' in writer
    assert 'y' in writer.x
    assert 'x/y/z' in writer
    assert 'z' in writer.x.y
    assert 'd' in writer.a.b.c
    assert '/a/d3' in writer
    assert 'd4' in writer
    assert '/d5' in writer.a.b.c.d
    assert '/a/b/c/d/d6' in writer
    assert 'd7' in writer
    assert writer.a is not root.a
    assert writer.a.d1 is not root.a.d1
    assert writer.a.b is not root.a.b
    assert writer.a.b.d2 is not root.a.b.d2
    assert writer.a.b.c is not root.a.b.c
    assert writer.a.b.c.d is not root.a.b.c.d
    assert writer.a.b.c.d.d5 is not root.a.b.c.d.d5
    assert writer.a.b.c.d.d6 is not root.a.b.c.d.d6
    assert writer.a.d3 is not root.a.d3
    assert writer.d4 is not root.d4
    assert writer.d7 is not root.d7
    assert writer.x is not root.x
    assert writer.x.y is not root.x.y
    assert writer.x.y.z is not root.x.y.z
    assert len(writer.metadata) == 4
    assert writer.metadata.one == 1
    assert writer.metadata.foo == 'bar'
    assert writer.metadata.two == 2
    assert writer.metadata.three == 3

    # d4 is a Dataset
    with pytest.raises(TypeError):
        writer.set_root(root.d4)

    # pass in a Group instead of a Root
    writer.set_root(root.x)
    assert len(writer) == 2
    assert len(list(writer.groups())) == 2
    assert len(list(writer.datasets())) == 0
    assert 'x' not in writer
    assert 'y' in writer
    assert 'y/z' in writer
    assert 'z' in writer.y
    assert len(writer.metadata) == 0
